! *************************************************************
! MODULE MSUBFC
!  definition of fuel consumption calculating subroutine
! *************************************************************
module msubfc

    interface

        subroutine fcgrid( gfcm, fcm, ftqc )
            use hvtransfc
            type (sgridmap) gfcm
            type (sfcmap) fcm
            type (stqcurve) ftqc
        end subroutine fcgrid

        subroutine calcfc( rev, tq, sp, fc, n, ftqc, gfcm, eng )
            use hvtransfc
            type (sgridmap) gfcm
            type (stqcurve) ftqc
            type (sengine) eng                                          ! engine
            real(8) rev(:), tq(:), fc(:)
            integer n, sp(:)
        end subroutine calcfc
        
        subroutine calcistop(v, rev, fc, n, t_istop, t_istop2, t_istop_off)
            integer n, i, j, t_istop, t_istop2, t_istop_off
            integer flag_istop(99999)
            real(8) v(:), rev(:), fc(:)
        end subroutine calcistop

        subroutine calcave( v, fc, n, avv, avfc )
            integer n
            real(8) v(:), fc(:), avv, avfc
        end subroutine calcave

    end interface

end module msubfc



! *************************************************************
! SUBROUTINE CALCFC
!  calculation of fuel consumption at every time step
!  rev : engine speed array
!  tq  : engine torque array
!  sp : shift position array
!  fc : fuel consumption array (output)
!  n : number of data
!  ftqc : frictional torque
!  gfcm : grid fuel map
! *************************************************************
subroutine calcfc( rev, tq, sp, fc, n, ftqc, gfcm, eng )
    use hvtransfc
    use mshiftp, only : maxtqlin
    implicit none

    type ( sgridmap ) :: gfcm
    type ( stqcurve ) :: ftqc
    type (sengine) eng                                          ! engine
    real(8) :: rev(:), tq(:), fc(:)
    integer :: sp(:)
    integer n

    integer i, j, ierr
    real(8) ftq
    real(8), allocatable :: f(:), d2(:), d(:,:)
    logical skip

    allocate( d( gfcm%nx, gfcm%ny ) )
    allocate( f( gfcm%ny ) )
    allocate( d2( gfcm%ny ) )

    do j = 1, n
        if ( sp(j) == 0 ) then
            fc(j) = gfcm%fcidle                                 ! idling
            cycle
        end if

        if ( rev(j) == eng%nidle) then
            fc(j) = gfcm%fcidle                                 ! idling  add. 170118
            cycle
        end if

        ftq = maxtqlin( ftqc, rev(j) )
        if ( tq(j) <= ftq ) then
            fc(j) = 0.0_8
        else
            do i = 1, gfcm%ny
                call DPCHIM( gfcm%nx, gfcm%rev, gfcm%fc(1,i), d(1,i), 1, ierr )
                if( ierr < 0 ) then
                    call errhermite ( 5, ierr )
                end if

                skip = .true.
                call DPCHFE( gfcm%nx, gfcm%rev, gfcm%fc(1,i), d(1,i),   &
                             1, skip, 1, rev(j), f(i), ierr)    ! torque-fuel curve
                if( ierr < 0 ) then
                    call errhermite ( 6, ierr )
                end if
            end do

            skip=.true.
            call DPCHIM( gfcm%ny, gfcm%tq, f, d2, 1, ierr )
            if( ierr < 0 ) then
                call errhermite ( 7, ierr )
            end if

            call DPCHFE( gfcm%ny,gfcm%tq,f,d2,1,skip,1,tq(j),fc(j),ierr )   ! fc at tq(j)
            if( ierr < 0 ) then
                call errhermite ( 8, ierr )
            end if

            if( fc(j) < 0.0_8 ) then
                fc(j) = 0.0_8
            end if
        end if
    end do

    deallocate( d, d2,f )

    end subroutine


! *************************************************************
! SUBROUTINE calcistop
!   calculate ISS
!  v : calculated speed
!  fc : fuel consumption
!  rev: engine speed
!  n : number of data
!  t_istop: AChXgbv쓮҂
!  t_istop2: AChXgbv쓮҂ԁiJE05X^[gj
!  t_istop_off: ԗiENGn^C~O
! *************************************************************
subroutine calcistop(v, rev, fc, n, t_istop, t_istop2, t_istop_off)
    implicit none
    integer n, i, j, t_istop, t_istop2, t_istop_off
    real(8) v(:), rev(:), fc(:)
    integer flag_istop(99999)

    flag_istop = 0  ! 0:istop_off(engine_on),  1:istop_on(engine_off)
! JE05X^[gȊO
    do i=25, n        ! @ŏ25b̓XLbviJE05pj
        if(v(i)==0.0)then
            flag_istop(i) = 1  ! UISS_ON
            do j=i, min(n, i + t_istop_off)           ! t_istop_offOɂISS_OFF
                if(v(j).ne.0.0)then
                    flag_istop(i) = 0
                    exit
                end if
            end do
            do j=i, max(1,i - t_istop), -1           ! ~t_istopISS_ON
                if(v(j).ne.0.0)then
                    flag_istop(i) = 0
                    exit
                end if
            end do  
        end if
    end do
! JE05X^[g
    do i=1, 25        ! @JE05X^[g
        flag_istop(i) = 0  ! UISS_OFF
        if(i > t_istop2)then
            flag_istop(i) = 1 ! ISS_ON
            do j=i, min(n, i + t_istop_off)
                if(v(j).ne.0.0)then
                    flag_istop(i) = 0 ! ISS_OFF
                    exit
                end if
            end do
        end if
    end do
    
    do i=1, n
        if(flag_istop(i)==1) then
            fc(i) = 0.0
            rev(i) = 0.0
        end if
    end do   
                

end subroutine  
    


! *************************************************************
! SUBROUTINE CALCAVE
!  calculate average speed, average fuel consumption
!  v : calculated speed
!  fc : fuel consumption
!  n : number of data
!  avv : average speed
!  avfc : average fuel consumption
!  t_istop : time between vehicle stop and start of idling-stop
!  t_istop2 : istop_start_time after JE05 start 
! *************************************************************
subroutine calcave( v, fc, n, avv, avfc)
    implicit none

    integer n, i, j
    real(8) v(:), fc(:)
    real(8) avv, avfc, sumv, sumf

    sumv = 0.0_8
    sumf = 0.0_8
    do i = 1, n
        sumv = sumv + v(i)
        sumf = sumf + fc(i)
    end do
 
    avv = sumv / (real(n,8) * 1.0_8 )
    avfc = sumv / sumf

end subroutine


! *************************************************************
! SUBROUTINE FCGRID
!  generate NXxNY grid fuel map
!  gfcm   : type(sgridmap), grid fuel map
!  fcm    : type(sfcmap), input fuel map
!  ftqc   : type(stqcurve), frictional torque
! *************************************************************
subroutine fcgrid( gfcm, fcm, ftqc )
    use hvtransfc
    use mshiftp, only : maxtqlin
    implicit none

    integer, parameter :: nx = 20, ny = 20
    type (stqcurve) :: ftqc
    type (sfcmap) :: fcm
    type (sgridmap) :: gfcm

    integer i, j, ierr
    integer rn, tn, tnmax
    real(8), allocatable :: r(:), t(:,:), f(:,:), x(:)
    integer, allocatable :: tqn(:)
    real(8) nemin, nemax, tqmin, tqmax
    real(8) dne, dtq
    real(8) rr
    logical skip
    real(8), allocatable :: d1(:,:), f1(:,:)
    real(8), pointer :: pchval(:)

    allocate ( gfcm%rev( nx ) )
    allocate ( gfcm%tq ( ny ) )
    allocate ( gfcm%fc ( nx, ny ) )
    gfcm%nx = nx
    gfcm%ny = ny
    gfcm%fcidle = fcm%fcidle

    nemin = minval( fcm%rev )
    nemax = maxval( fcm%rev )
    tqmin = minval( fcm%tq )
    tqmax = maxval( fcm%tq )

    rn = 1                                                      ! count points of input fuel map
    tn = 1
    tnmax = 1
    rr = fcm%rev(1)
    do i = 1, fcm%ndata
        if ( rr /= fcm%rev(i) ) then
            rn = rn + 1
            rr = fcm%rev(i)
            if( tn > tnmax ) tnmax = tn
            tn = 1
        end if
        tn = tn + 1
    end do

    allocate ( r(rn) )
    allocate ( tqn(rn) )
    allocate ( t(rn, tnmax) )
    allocate ( f(rn, tnmax) )

    j = 1
    r(j) = fcm%rev(1)
    tqn(j) = 1
    t( j, tqn(j) ) = maxtqlin( ftqc, r(j) )
    f( j, tqn(j) ) = 0.0_8
    do i = 1, fcm%ndata
        if ( r(j) /= fcm%rev(i) ) then
            j = j + 1
            r(j) = fcm%rev(i)
            tqn(j) = 1
            t( j, tqn(j) ) = maxtqlin( ftqc, r(j) )
            f( j, tqn(j) ) = 0.0_8
            tqmin = min( tqmin, t(j,tqn(j)) )
        end if
        tqn(j) = tqn(j) + 1
        t( j, tqn(j) ) = fcm%tq(i)
        f( j, tqn(j) ) = fcm%fc(i)
    end do

!   -----------------------------------------------------------
    dne = ( nemax - nemin ) / real(nx-1, 8)
    dtq = ( tqmax - tqmin ) / real(ny-1, 8)

    do i = 1, nx
        gfcm%rev(i) = nemin + dne * real(i-1, 8)
    end do
    do i = 1, ny
        gfcm%tq(i) = tqmin + dtq * real(i-1, 8)
    end do

!   -----------------------------------------------------------
    allocate ( f1(rn, ny) )
    allocate ( d1(rn, tnmax) )
    do i = 1, rn
        allocate( x( tqn(i) ) )
        do j = 1, tqn(i)
            x(j) = t( i, j )
        end do

        call DPCHIM( tqn(i), x, f(i,1), d1(i,1), rn, ierr )
        if( ierr < 0 ) call errhermite ( 1, ierr )

        skip = .true.
        allocate ( pchval( ny ) )
        call DPCHFE( tqn(i), x, f(i,1), d1(i,1), rn, skip,          &
                     gfcm%ny, gfcm%tq, pchval, ierr )
        if( ierr < 0 ) call errhermite ( 2, ierr )

        do j = 1, ny
            f1(i,j) = pchval(j)
        end do

        deallocate ( pchval, x )
    end do

    deallocate ( d1 )

!   -----------------------------------------------------------
    allocate ( d1(rn, ny) )
    do i = 1, ny
        call DPCHIM( rn, r, f1(1,i), d1(1,i), 1, ierr )
        if( ierr < 0 ) call errhermite ( 3, ierr )

        skip = .true.
        allocate ( pchval( nx ) )
        call DPCHFE( rn, r, f1(1,i), d1(1,i), 1, skip,          &
                     gfcm%nx, gfcm%rev, pchval, ierr )
        if( ierr < 0 ) call errhermite ( 4, ierr )

        do j = 1, nx
            gfcm%fc(j,i) = pchval(j)
        end do

        deallocate ( pchval )
    end do

    deallocate ( d1, f1, r, tqn, t, f )

end subroutine




! *************************************************************
! SUBROUTINE ERRHERMITE
!  error handler for hermite interpolation
! *************************************************************
subroutine errhermite ( errid, ierr )
    implicit none

    integer errid, ierr
    character (len=1024) msg

    select case ( errid )
!   -----------------------------------------------------------
!   at FCGRID
!   -----------------------------------------------------------
    case( 1 )
        msg = "Error : DPCHIM ( called from FCGRID, 1st )"
    case( 2 )
        msg = "Error : DPCHFE ( called from FCGRID, 1st )"
    case( 3 )
        msg = "Error : DPCHIM ( called from FCGRID, 2nd )"
    case( 4 )
        msg = "Error : DPCHFE ( called from FCGRID, 2nd )"

!   -----------------------------------------------------------
!  at CALCFC
!   -----------------------------------------------------------
    case( 5 )
        msg = "Error : DPCHIM ( called from CALCFC, 1st )"
    case( 6 )
        msg = "Error : DPCHFE ( called from CALCFC, 1st )"
    case( 7 )
        msg = "Error : DPCHIM ( called from CALCFC, 2nd )"
    case( 8 )
        msg = "Error : DPCHFE ( called from CALCFC, 2nd )"
    end select

    write( 0, '( A )' ) msg
    write ( 0, '( A, I )' ) "ierr = ", ierr
    stop ' Terminated by error of Hermite function. Check input fuel map.'

end subroutine



! **************************************************************
! DPCHST:  SIGN TESTING OF 2 ARGUMENTS
!     Returns:
!        -1. if ARG1 and ARG2 are of opposite sign.
!         0. if either argument is zero.
!        +1. if ARG1 and ARG2 are of the same sign.
! **************************************************************
real(8) function DPCHST( arg1, arg2 )
    implicit none

    real(8) arg1, arg2

    if ( ( arg1 == 0.0_8 ) .or. ( arg2 == 0.0_8 ) ) then
        DPCHST = 0.0_8
    else
        DPCHST = dsign(1.0_8, arg1) * dsign(1.0_8, arg2)
    end if

end function



! **************************************************************
! DPCHIM: Piecewise Cubic Hermite Interpolation to Monotone data
! **************************************************************
subroutine DPCHIM ( n, x, f, d, incfd, ierr )
    implicit none
    interface
        real(8) function DPCHST( arg1, arg2 )
            real(8) arg1, arg2
        end function
    end interface

    integer n, incfd, ierr
    real(8) x(*), f(incfd,*), d(incfd,*)

    integer i, nsec
    real(8) dx1, dx2, del1, del2, dsave, dxsum, dxsumt3, dmax, dmin
    real(8) w1,w2, drat1, drat2, sgn

!   ----- check argument ---------------------------------------
    ierr = 0
    if( n < 2 ) then
        ierr = -1
        return
    end if

    if( incfd < 1 ) then
        ierr = -2
        return
    end if

    do i = 2, n
        if( x(i) > x(i-1) ) then
            continue
        else
            ierr = -3
            return
        end if
    end do

!   ----- Piecewise Cubic Hermite Interpolation ----------------
    nsec = n -1
    dx1 = x(2) - x(1)
    del1 = ( f(1,2) - f(1,1) ) / dx1
    dsave = del1

    if( n == 2 )then
        d(1,1) = del1
        d(1,2) = del1
        return
    end if

    dx2 = x(3) - x(2)
    del2 = ( f(1,3) - f(1,2) ) / dx2

    dxsum = dx1 + dx2
    w1 = ( dx1 + dxsum ) / dxsum
    w2 = -dx1 / dxsum
    d(1,1) = w1 * del1 + w2 * del2
    if( DPCHST( d(1,1), del1 ) <= 0.0_8 ) then
        d(1,1) = 0.0_8
    else if( DPCHST( del1, del2 ) < 0.0_8 ) then
        dmax = 3.0_8 * del1
        if( abs( d(1,1) ) > abs( dmax ) ) then
            d(1,1) = dmax
        end if
    end if

    do i = 2, nsec
        if( i /= 2 )then
            dx1 = dx2
            dx2 = x(i+1) - x(i)
            dxsum = dx1 + dx2
            del1 = del2
            del2 = ( f(1,i+1) - f(1,i) ) / dx2
        end if

        d( 1, i ) = 0.0_8
        sgn = DPCHST( del1, del2 )

        if( sgn == -1.0_8 ) then
            ierr = ierr + 1
            dsave = del2

        else if( sgn == 0.0_8 ) then
            if( del2 /= 0.0_8 ) then
                if( DPCHST( dsave, del2 ) < 0.0_8 ) then
                    ierr = ierr + 1
                end if
                dsave = del2
            end if

        else
            dxsumt3 = dxsum * 3.0_8
            w1 = ( dxsum + dx1 ) / dxsumt3
            w2 = ( dxsum + dx2 ) / dxsumt3
            dmax = max( abs(del1), abs(del2) )
            dmin = min( abs(del1), abs(del2) )
            drat1 = del1 / dmax
            drat2 = del2 / dmax
            d(1,i) = dmin / ( w1 * drat1 + w2 * drat2 )
        end if
    end do

    w1 = -dx2 / dxsum
    w2 = ( dx2 + dxsum ) / dxsum
    d( 1, n ) = w1 * del1 + w2 * del2
    if( DPCHST( d(1,n), del2 ) <= 0.0_8 ) then
        d(1,n) = 0.0_8
    else if( DPCHST( del1, del2 ) < 0.0_8 ) then
        dmax = 3.0_8 * del2
        if( abs( d(1,n) ) > abs( dmax ) ) then
            d( 1, n ) = dmax
        end if
    end if

end subroutine DPCHIM



! **************************************************************
! DPCHFE:  Piecewise Cubic Hermite Function Evaluator
! **************************************************************
subroutine DPCHFE( n, x, f, d, incfd, skip, ne, xe, fe, ierr )
    implicit none

    integer n                                                   ! f[^_
    integer incfd, ne, ierr
    real(8) x(*), f(incfd,*), d(incfd,*), xe(*), fe(*)
    logical skip

    integer i, ierc, ir, j, j2,jfirst, next(2), nj

    ierr = 0
!   ----- check argument ---------------------------------------
    if( skip ) then
        if( n < 2 ) then
            ierr = -1
            return
        end if

        if( incfd < 1 ) then
            ierr = -2
            return
        end if

        do i = 2, n
            if( x(i) <= x(i-1) ) then
                ierr = -3
                return
            end if
        end do

        if( ne < 1 ) then
            ierr = -4
            return
        end if
    end if


!   ----- piecewise cubic hermite function evaluator ------------
    skip = .true.
    jfirst = 1
    ir = 2

    do while ( jfirst <= ne )
        j2 = ne + 1
        do j = jfirst, ne
            if( xe(j) >= x(ir) ) then
                if( ir == n ) then
                    j2 = ne + 1
                else
                    j2 = j
                end if
                exit
            end if
        end do
        nj = j2 - jfirst

        if( nj /= 0 ) then
            call DCHFEV ( x(ir-1), x(ir), f(1,ir-1), f(1,ir),       &
                          d(1,ir-1), d(1,ir), nj, xe(jfirst),       &
                          fe(jfirst), next, ierc )
            if( ierc < 0 ) then
                ierr = -5
                return
            end if

            if( next(2) == 0 ) then

            else
                if( ir < n ) then
                    ierr = -5
                    return
                else
                    ierr = ierr + next(2)
                end if
            end if

            if( next(1) == 0 ) then

            else
                if( ir > 2 ) then
                    do i = jfirst, j2-1
                        if( xe(i) < x(ir-1) ) then
                            exit
                        end if
                    end do
                    if( i > j2-1 ) then
                        ierr = -5
                        return
                    end if
                    j2 = i
                    do i = 1, ir-1
                        if( xe(j2) < x(i) ) exit
                    end do
                    ir = max( 1, i-1 )
                else
                    ierr = ierr + next(1)
                end if
            end if
            jfirst = j
        end if

        ir = ir + 1
        if( ir > n ) exit
    end do

end subroutine



! **************************************************************
! DCHFEV:  Cubic Hermite Function EValuator
! **************************************************************
subroutine DCHFEV (x1, x2, f1, f2, d1, d2, ne, xe, fe, next, ierr)
    implicit none

    integer ne, ierr, next(2)
    real(8) x1, x2, f1, f2, d1, d2, xe(*), fe(*)
    integer i
    real(8) c2, c3, del1, del2, delta,h, x, xmi, xma

    ierr = 0
!   ----- check arguments---------------------------------------
    if( ne < 1 ) then
        ierr = -1
        return
    end if

    h = x2 - x1
    if( h == 0.0_8 ) then
        ierr = -2
        return
    end if

!   ----- cubic hermite function evaluator ---------------------
    next(1) = 0
    next(2) = 0
    xmi = min( 0.0_8, h )
    xma = max( 0.0_8, h )

    delta = ( f2 - f1 ) / h
    del1 = ( d1 - delta ) / h
    del2 = ( d2 - delta ) / h
    c2 = -( del1 + del1 + del2 )
    c3 = ( del1 + del2 ) / h

    do i = 1, ne
        x = xe(i) - x1
        fe(i) = f1 + x * ( d1 + x * ( c2 + x * c3 ) )
        if( x < xmi ) then
            next(1) = next(1) + 1
        end if
        if( x > xma ) then
            next(2) = next(2) + 1
        end if
    end do

end subroutine
